import textwrap
import operator
import json
from collections import defaultdict
from typing_extensions import TypedDict
from langchain_core.messages import HumanMessage, SystemMessage
from typing import Annotated, Sequence
from langgraph.graph import StateGraph, START, END
from agentiacap.tools.document_intelligence import process_base64_files, ImageFieldExtractor
from agentiacap.configs.classes import Input
from agentiacap.configs.llms import llm4o
from agentiacap.configs.Prompt_Template import TextExtractorPrompt, fields_to_extract, merger_definition
from agentiacap.tools.convert_pdf import pdf_base64_to_image_base64

def print_truncated(data, max_length=50, indent=0):
    """Imprime un diccionario truncando los strings largos sin modificar los datos originales."""
    if isinstance(data, dict):
        for key, value in data.items():
            print(" " * indent + str(key) + ":")
            print_truncated(value, max_length, indent + 2)
    elif isinstance(data, list):
        for item in data:
            print_truncated(item, max_length, indent)
    elif isinstance(data, str):
        print(" " * indent + textwrap.shorten(data, width=max_length, placeholder="..."))
    else:
        print(" " * indent + str(data))

def find_missing_fields(data):
    """
    Busca recursivamente una clave llamada 'missing_fields' dentro de una estructura arbitraria 
    de listas y diccionarios, y devuelve todas las listas encontradas.
    
    :param data: Puede ser un dict o list con estructuras anidadas desconocidas.
    :return: Lista con todos los valores encontrados bajo la clave 'missing_fields'.
    """
    results = []

    if isinstance(data, dict):
        for key, value in data.items():
            if key == "missing_fields" and isinstance(value, list):
                results.append(value)
            else:
                results.extend(find_missing_fields(value))

    elif isinstance(data, list):
        for item in data:
            results.extend(find_missing_fields(item))

    return results

class ResultExtraction(TypedDict):
    fuente:Annotated[str, ...]
    valores:Annotated[list, ...]

class OutputState(TypedDict):
    extractions:Annotated[list, operator.add]
    tokens:Annotated[int, ...]

merger = merger_definition | llm4o.with_structured_output(ResultExtraction)

class State(TypedDict):
    aggregate: Annotated[list, operator.add]
    tokens: int
    text: str   # Almacena asunto y cuerpo del mail
    images: list  # Almacena las imagenes adjuntas
    pdfs: list  # Almacena los pdfs adjuntos

class Fields(TypedDict):
    customer_name:str
    customer_tax_id:str
    invoice_id:str
    vendor_tax_id:str
    purchase_order_number:str
    invoice_date:str
    invoice_total:str


class ClassifyNode:
    def __call__(self, state:Input) -> State:
        image_extensions = (".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".webp")
        pdf_extension = (".pdf")
        images, pdfs = [], []
        files = state["adjuntos"]
        for file in files:
            file_name = file.get("file_name", "").lower()
            if file_name.endswith(image_extensions):
                print(f"DEBUG - Clasificando como imagen: {file_name}")
                images.append(file)
            elif file_name.endswith(pdf_extension):
                print(f"DEBUG - Clasificando como PDF: {file_name}")
                pdfs.append(file)

        return {"images": images, "pdfs": pdfs, "text": str(state["asunto"]) + str(state["cuerpo"]), "tokens":0}


class VisionNode:
    async def __call__(self, state: State) -> State:
        images_from_pdfs = []
        for file in state["pdfs"]:
            pages = pdf_base64_to_image_base64(file["base64_content"], 1)
            for page in pages:
                image = {
                    "file_name": file["file_name"],
                    "base64_content": page
                }
                images_from_pdfs.append(image)
        extractor = ImageFieldExtractor()
        result = extractor.extract_fields(base64_images=state["images"]+images_from_pdfs, fields_to_extract=fields_to_extract)
        tokens = 0
        for element in result:
            tokens += int(result[element]["tokens"])
        print(f"DEBUG - Extraccion de Vision: {result}")
        return {"tokens": state["tokens"] + tokens, "aggregate": [result]}

class ImageNode:
    async def __call__(self, state: State) -> State:
        extractor = ImageFieldExtractor()
        result = extractor.extract_fields(base64_images=state["images"], fields_to_extract=fields_to_extract)
        tokens = 0
        for element in result:
            tokens += int(result[element]["tokens"])
        print(f"DEBUG - Extraccion de imagenes: {result}")
        return {"tokens": state["tokens"] + tokens, "aggregate": [result]}

class PrebuiltNode():
    async def __call__(self, state: State) -> State:
        result = process_base64_files(base64_files=state["pdfs"], fields_to_extract=fields_to_extract)
        return {"aggregate": [result]}

class NamesAndCuitsNode:
    async def __call__(self, state:State) -> Fields:
        prompt = [SystemMessage(content=TextExtractorPrompt.names_and_cuits_prompt)] + [HumanMessage(content=f"Dado el siguiente texto de un mail extrae el dato pedido: {state['text']}")]
        result = await llm4o.ainvoke(prompt)
        content = result.content.strip("```json").strip("```")  # Limpia los delimitadores
        data = json.loads(content)
        return {"customer_name": data["CustomerName"], "customer_tax_id": data["CustomerTaxId"], "vendor_tax_id": data["VendorTaxId"]}
    
class InvoiceNode:
    async def __call__(self, state:State) -> Fields:
        prompt = [SystemMessage(content=TextExtractorPrompt.invoice_id_prompt)] + [HumanMessage(content=f"Dado el siguiente texto de un mail extrae los datos pedidos: {state['text']}")]
        result = await llm4o.ainvoke(prompt)
        content = result.content.strip("```json").strip("```")  # Limpia los delimitadores
        content = json.loads(content)
        print(f"Invoice data: {content}")
        return {"invoice_id": content["InvoiceId"], "invoice_date": content["InvoiceDate"], "invoice_total": content["InvoiceTotal"]}

class MergeFieldsNode:
    async def __call__(self, state: Fields) -> State:
        missing_fields = []
        for field in fields_to_extract:
            if field not in state:
                missing_fields.append(field)
        result = {
            "Mail":{
                "page_number": 1,
                "fields":state, 
                "missing_fields":missing_fields, 
                "error":"",
                "source": "Mail"
            },
        }
        # Mantenemos el estado original y solo agregamos "aggregate"
        return {"aggregate": [result]}

# Analizo todo los adjuntos si los hay
def router(state: State) -> Sequence[str]:
    routes = []
    if state["images"]:
        print("DEBUG - router GO TO vision")
        routes.append("extract from images")
    
    if state["pdfs"]:
        print("DEBUG - router GO TO prebuilt")
        routes.append("extract with prebuilt")

    if len(routes) == 0:
        print("DEBUG - router GO TO merger")
        return ["merger"]
    
    return routes

def merge_results(state: State) -> OutputState:
    # Resultado de agrupación
    grouped_data = defaultdict(list)

    # Procesar los datos
    for item in state["aggregate"]:
        if isinstance(item, list):  # Caso: Lista de documentos
            print("instancia de lista")
            for doc in item:
                source = doc['source']
                grouped_data[source].append(doc)
        elif isinstance(item, dict):  # Caso: Diccionario de archivos
            print("instancia de diccionario")
            for key, value in item.items():
                source = value['source']
                # Eliminar el campo 'source'
                value.pop('source', None)
                grouped_data[source].append(value)

    # Formatear la salida
    formatted_data = []
    for source, values in grouped_data.items():
        formatted_data.append({
            'fuente': source,
            'valores': values
        })
    return {"extractions":formatted_data, "tokens":state["tokens"]}

def should_continue(state:State):
    print("DEBUG - GO TO Vision Node")
    return "vision"

# Construcción del grafo
builder = StateGraph(State, input=Input, output=OutputState)

builder.add_node("initializer", ClassifyNode())
builder.add_node("extract names and cuits", NamesAndCuitsNode())
builder.add_node("extract invoices IDs", InvoiceNode())
builder.add_node("merge fields", MergeFieldsNode())
builder.add_node("extract from images", ImageNode())
builder.add_node("extract with vision", VisionNode())
builder.add_node("extract with prebuilt", PrebuiltNode())
builder.add_node("merger", merge_results)

builder.add_edge(START, "initializer")
builder.add_edge("initializer", "extract names and cuits")
builder.add_edge("initializer", "extract invoices IDs")
builder.add_edge("extract invoices IDs", "merge fields")
builder.add_edge("extract names and cuits", "merge fields")
builder.add_conditional_edges("initializer", router, ["extract with prebuilt", "extract from images", "merger"])
builder.add_conditional_edges("extract with prebuilt", should_continue, {"vision":"extract with vision", END:"merger"})
builder.add_edge("extract from images", "merger")
builder.add_edge("extract with vision", "merger")
builder.add_edge("merge fields", "merger")
builder.add_edge("merger", END)

extractor = builder.compile()


import os
from dotenv import load_dotenv
from langchain_openai import AzureChatOpenAI

# Cargar variables de entorno desde el archivo .env
load_dotenv()

llm4o_mini = AzureChatOpenAI(
    azure_deployment="gpt-4o-mini",  
    api_version="2024-02-15-preview",
    temperature=0,
    max_tokens=10000,
    timeout=None,
    max_retries=2
)

llm4o = AzureChatOpenAI(
    azure_deployment="gpt-4o",  
    api_version="2024-08-01-preview",
    temperature=0,
    max_tokens=10000,
    timeout=None,
    max_retries=2
)


import base64
import io
import fitz
from PIL import Image
from pdf2image import convert_from_path

def pdf_page_to_image(pdf_path: str, page_number: int):
    try:
        images = convert_from_path(pdf_path, first_page=page_number, last_page=page_number)
        if images:
            print(f"✅ Página {page_number} convertida a imagen correctamente.")
            return images[0]  # Retorna la imagen de la página
        else:
            print(f"⚠️ No se pudo extraer la imagen de la página {page_number}, intentando renderizar con PyMuPDF.")
            return render_pdf_page_as_image(pdf_path, page_number)
    except Exception as e:
        print(f"❌ Error al convertir la página {page_number} a imagen: {e}")
        return None

def render_pdf_page_as_image(pdf_path: str, page_number: int):
    try:
        doc = fitz.open(pdf_path)
        page = doc[page_number - 1]  # Índice basado en 1
        pix = page.get_pixmap()
        img = Image.frombytes("RGB", [pix.width, pix.height], pix.samples)
        print(f"✅ Página {page_number} renderizada con PyMuPDF correctamente.")
        return img
    except Exception as e:
        print(f"❌ Error al renderizar la página {page_number} con PyMuPDF: {e}")
        return None

def pdf_page_to_base64(pdf_path: str, page_number: int):
    image = pdf_page_to_image(pdf_path, page_number)
    if image is None:
        print(f"⚠️ No se pudo convertir la página {page_number} a base64 porque la imagen es inválida.")
        return None
    
    try:
        buffer = io.BytesIO()
        image.save(buffer, format="PNG")
        base64_string = base64.b64encode(buffer.getvalue()).decode("utf-8")
        print(f"✅ Página {page_number} convertida a base64 correctamente.")
        return base64_string
    except Exception as e:
        print(f"❌ Error al convertir la página {page_number} a base64: {e}")
        return None

def pdf_base64_to_image(pdf_base64: str, page_number: int):
    try:
        pdf_bytes = base64.b64decode(pdf_base64)
        pdf_stream = io.BytesIO(pdf_bytes)
        doc = fitz.open(stream=pdf_stream, filetype="pdf")
        if page_number <= len(doc):
            page = doc[page_number - 1]  # Índice basado en 1
            pix = page.get_pixmap()
            img = Image.frombytes("RGB", [pix.width, pix.height], pix.samples)
            print(f"✅ Página {page_number} convertida desde base64 a imagen correctamente.")
            return img
        else:
            return None
    except Exception as e:
        print(f"❌ Error al convertir la página {page_number} desde base64 a imagen: {e}")
        return None

def pdf_base64_to_image_base64(pdf_base64: str, fin: int):
    conversiones = []
    for page_number in range(fin):
        image = pdf_base64_to_image(pdf_base64, page_number)
        if image is None:
            print(f"⚠️ No se pudo convertir la página {page_number} a base64 porque la imagen es inválida o se alcanzo el fin de archivo.")
            break

        try:
            buffer = io.BytesIO()
            image.save(buffer, format="PNG")
            base64_string = base64.b64encode(buffer.getvalue()).decode("utf-8")
            print(f"✅ Página {page_number} convertida desde base64 a base64 correctamente.")
            conversiones.append(base64_string)
        except Exception as e:
            print(f"❌ Error al convertir la página {page_number} desde pdf base64 a imagen base64: {e}")
            break
    return conversiones


from typing_extensions import Annotated, TypedDict

class Mail(TypedDict):
    asunto:Annotated[str, ...]
    cuerpo:Annotated[str, ...]
    adjuntos:Annotated[list, ...]
    categoria:Annotated[str, ...]
    extracciones:Annotated[list, ...]
    tokens:Annotated[int, ...]
    
class Result(TypedDict):
    category:Annotated[str, ...]
    extractions:Annotated[list, ...]
    tokens:Annotated[int, ...]

# Schemas de entrada y salida
class Input(TypedDict):
    asunto:Annotated[str, ...]
    cuerpo:Annotated[str, ...]
    adjuntos:Annotated[list, ...]

class Output(TypedDict):
    result: Annotated[dict, ...]


from typing import Annotated, TypedDict
from langchain_core.messages import HumanMessage
from langgraph.graph import StateGraph, START, END
from agentiacap.configs.Prompt_Template import cleaner_definition
from agentiacap.configs.llms import llm4o
from agentiacap.configs.classes import Input

class Cuerpo(TypedDict):
    cuerpo:Annotated[str, ...]

clean = cleaner_definition | llm4o.with_structured_output(Cuerpo)

# Defino nodes
def clean_body(state: Input) -> Input:
    cuerpo_filtrado = clean.invoke([HumanMessage(
        content = f"""Limpia el siguiente mail:\n{state["cuerpo"]}""")])
    return {"cuerpo":cuerpo_filtrado}

def clean_attachments(state: Input) -> Input:
    if len(state["adjuntos"]) == 0:
        return state
    
    return state

builder = StateGraph(input=Input, output=Input)

builder.add_node("Clean body", clean_body)
builder.add_node("Clean attachments", clean_attachments)

builder.add_edge(START, "Clean body")
builder.add_edge("Clean body", "Clean attachments")
builder.add_edge("Clean attachments", END)

cleaner = builder.compile()



from typing import Literal
from langgraph.types import Command
from langgraph.graph import StateGraph, START, END
from agentiacap.agents.agentCleaner import cleaner
from agentiacap.agents.agentClassifier import classifier
from agentiacap.agents.agentExtractor import extractor
from agentiacap.configs.classes import Input, Output, Mail
import logging

logging.basicConfig(level=logging.ERROR)  # Puedes usar INFO, WARNING, etc.
logger = logging.getLogger("langchain")
logger.setLevel(logging.ERROR)

relevant_categories = ["Estado de facturas", "Pedido devolución retenciones", "Impresión de OP y/o Retenciones"]

async def call_cleaner(state: Input) -> Mail:
    cleaned_result = await cleaner.ainvoke(state)
    return {"asunto":cleaned_result["asunto"], "cuerpo":cleaned_result["cuerpo"], "adjuntos":cleaned_result["adjuntos"]}

async def call_classifier(state: Mail) -> Command[Literal["Extractor", "Output"]]:
    input_schema = Input(asunto=state["asunto"], cuerpo=state["cuerpo"], adjuntos=state["adjuntos"])
    classified_result = await classifier.ainvoke(input_schema)
    if classified_result["category"] in relevant_categories:
        goto = "Extractor"
    else:
        goto = "Output"
    return Command(
        update={"categoria": classified_result["category"]},
        goto=goto
    )

async def call_extractor(state: Mail) -> Mail:
    input_schema = Input(asunto=state["asunto"], cuerpo=state["cuerpo"], adjuntos=state["adjuntos"])
    extracted_result = await extractor.ainvoke(input_schema)
    return {"extracciones": extracted_result["extractions"], "tokens": extracted_result["tokens"]}

def output_node(state: Mail) -> Output:
    result = {
        "category": state["categoria"],
        "extractions": state["extracciones"],
        "tokens": state["tokens"]
    }
    return {"result": result}

# Workflow principal
builder = StateGraph(Mail, input=Input, output=Output)

builder.add_node("Cleaner", call_cleaner)
builder.add_node("Classifier", call_classifier)
builder.add_node("Extractor", call_extractor)
builder.add_node("Output", output_node)

builder.add_edge(START, "Cleaner")
builder.add_edge("Cleaner", "Classifier")
builder.add_edge("Extractor", "Output")
builder.add_edge("Output", END)


graph = builder.compile()


import base64
import os
import json
from azure.core.credentials import AzureKeyCredential
from azure.ai.documentintelligence import DocumentIntelligenceClient
from azure.ai.documentintelligence.models import AnalyzeDocumentRequest
from openai import AzureOpenAI
from typing import List
from dotenv import load_dotenv

#   TODO: EL JSON DEVUELTO POR PROCESS_BASE_64_FILES NO CONTIENE MISSING FIELDS EN SU ESTRUCTURA

# Cargar las variables de entorno desde el archivo .env
load_dotenv()


def initialize_client():
    endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
    key = os.getenv("AZURE_OPENAI_API_KEY")
    return DocumentIntelligenceClient(endpoint=endpoint, credential=AzureKeyCredential(key))

def analyze_document_prebuilt_invoice(client, file_bytes: bytes, fields_to_extract: list) -> dict:
    data = {
        "results": [],
        "missing_fields": [],
        "error": ""
    }
    try:
        poller = client.begin_analyze_document(
            "prebuilt-invoice", AnalyzeDocumentRequest(bytes_source=file_bytes)
        )

        invoices = poller.result()

        if invoices.documents:
            for idx, invoice in enumerate(invoices.documents):
                invoice_data = []
                for field in fields_to_extract:
                    field_data = invoice.fields.get(field)
                    if field_data:
                        invoice_data.append({
                            field: field_data.content,
                        })
                    else:
                        data["missing_fields"].append(field)
                        invoice_data.append({
                            field: "none",
                        })

                data["results"].append({
                    "invoice_number": idx + 1,
                    "fields": invoice_data
                })
        else:
            data["error"] = "No se encontraron documentos en el archivo."

    except Exception as e:
        data["error"] = str(e)
    return data

def process_base64_files(base64_files: list, fields_to_extract: list) -> list:
    client = initialize_client()
    final_results = []

    for file_data in base64_files:
        file_name = file_data.get("file_name", "unknown")
        base64_content = file_data.get("base64_content", "")

        try:
            file_bytes = base64.b64decode(base64_content)
            print("DEBUG - Analizando PDF - ", file_name)
            # Intentar con text-based primero
            text_result = analyze_document_prebuilt_invoice(client, file_bytes, fields_to_extract)

            final_results.append({
                "file_name": file_name,
                "fields": text_result["results"],
                "missing_fields": text_result["missing_fields"],
                "error": text_result["error"],
                "source": "Document Intelligence"
            })

        except Exception as e:
            final_results.append({
                "file_name": file_name,
                "fields": [],
                "missing_fields": [],
                "error": str(e),
                "source": "Document Intelligence"
            })

    return final_results


class ImageFieldExtractor:
    def __init__(self):
        """
        Inicializa el cliente de OpenAI en Azure.
        :param openai_endpoint: Endpoint de Azure OpenAI.
        :param gpt_model_name: Nombre del modelo GPT configurado en Azure.
        :param api_key: Clave de la API para autenticación.
        :param api_version: Versión de la API de Azure OpenAI.
        """
        self.openai_client = AzureOpenAI(
            azure_endpoint=os.getenv("OPENAI_ENDPOINT"),
            api_key=os.getenv("OPENAI_API_KEY"),  # Usamos la API key para autenticación
            api_version=os.getenv("AZURE_OPENAI_API_VERSION"),
        )
        self.gpt_model_name = os.getenv("AZURE_OPENAI_DEPLOYMENT_NAME", "gpt-4o")

    def create_user_content(self, base64_data: str, fields_to_extract: List[str]):
        """
        Crea el contenido que se enviará al modelo para procesar.
        """
        user_content = [
            {
                "type": "text",
                "text": f"""
                    Extrae los siguientes campos del documento: {', '.join(fields_to_extract)}.
                    - Si un valor no está presente, indica null.
                    - Devuelve las fechas en formato dd-MM-yyyy.
                    - El "PurchaseOrderName" siempre es un número de 10 dígitos referenciado como orden de pago o similares y tiene la caracteristica de que siempre empieza con 2 o con 36. Ejemplos tipicos de este numero pueden ser 2000002154, 2000000831, 2000010953.  No siempre esta presente este dato.
                    -"CustomerName": se refiere a la sociedad por la que se hace la consulta. Solo se pueden incluir las sociedades permitidas en la lista de sociedades.
                    **Lista de sociedades permitidas:
                    -"Nombre Soc SAP": "AESA", "Código SAP": "0478", "Estado": "Activa", "CUIT": "30685211890", "Nombre en AFIP": "ASTRA EVANGELISTA SA"
                    -"Nombre Soc SAP": "YPF GAS", "Código SAP": "0522", "Estado": "Activa", "CUIT": "33555234649", "Nombre en AFIP": "YPF GAS S.A."
                    -"Nombre Soc SAP": "UTE LA VENTANA", "Código SAP": "0571", "Estado": "Activa", "CUIT": "30652671418", "Nombre en AFIP": "YACIMIENTO LA VENTANA YPF SA SINOPEC ARGENTINA EXPLORATION AND PRODUCTION INC UNION TRANSITORIA"
                    -"Nombre Soc SAP": "YPF S.A.", "Código SAP": "0620", "Estado": "Activa", "CUIT": "30546689979", "Nombre en AFIP": "YPF SA",
                    -"Nombre Soc SAP": "Fundación YPF", "Código SAP": "0789", "Estado": "Activa", "CUIT": "30691548054", "Nombre en AFIP": "FUNDACION YPF",
                    -"Nombre Soc SAP": "UTE LLANCANELO", "Código SAP": "0797", "Estado": "Activa", "CUIT": "30707293809", "Nombre en AFIP": "CONTRATO DE UNION TRANSITORIA DE EMPRESAS - AREA LLANCANELO U.T.E.",
                    -"Nombre Soc SAP": "OPESSA", "Código SAP": "0680", "Estado": "Activa", "CUIT": "30678774495", "Nombre en AFIP": "OPERADORAS DE ESTACIONES DE SERVICIO SA",
                    -"Nombre Soc SAP": "UTE CAMPAMENTO CENTRAL CAÑADON PERDIDO", "Código SAP": "0862", "Estado": "Activa", "CUIT": "33707856349", "Nombre en AFIP": "YPF S A - SIPETROL ARGENTINA S A - UTE CAMPAMENTO CENTRAL - CAÑADON PERDIDO",
                    -"Nombre Soc SAP": "UTE BANDURRIA", "Código SAP": "0900", "Estado": "Activa", "CUIT": "30708313587", "Nombre en AFIP": "YPF S.A WINTENSHALL ENERGIA SA - PAN AMERICAN ENERGY LLC AREA BANDURRIA UTE",
                    -"Nombre Soc SAP": "Ute Santo Domingo I y II", "Código SAP": "0901", "Estado": "Activa", "CUIT": "30713651504", "Nombre en AFIP": "GAS Y PETROELO DEL NEUQUEN SOCIEDAD ANONIMA CON PARTICIPACION ESTATAL MAYORITARIA - YPF S.A. - AREA SANTO DOMINGO I Y II UTE",
                    -"Nombre Soc SAP": "UTE CERRO LAS MINAS", "Código SAP": "0918", "Estado": "Activa", "CUIT": "30712188061", "Nombre en AFIP": "GAS Y PETROLEO DEL NEUQUEN SOCIEDAD ANONIMA CON PARTICIPACION ESTATAL MAYORITARIA-YPF S.A.-TOTAL AUSTRAL SA SUC ARG-ROVELLA ENERGIA SA-AREA CERRO LAS MINAS UTE",
                    -"Nombre Soc SAP": "UTE ZAMPAL OESTE", "Código SAP": "1046", "Estado": "Activa", "CUIT": "30709441945", "Nombre en AFIP": "YPF S.A EQUITABLE RESOURCES ARGENTINA COMPANY S.A - ZAMPAL OESTE UTE",
                    -"Nombre Soc SAP": "UTE ENARSA 1", "Código SAP": "1146", "Estado": "Activa", "CUIT": "30710916833", "Nombre en AFIP": "ENERGIA ARGENTINA S.A.- YPF S.A.- PETROBRAS ENERGIA S.A.- PETROURUGUAY S.A. UNION TRANSITORIAS DE EMPRESAS E1",
                    -"Nombre Soc SAP": "UTE GNL ESCOBAR", "Código SAP": "1153", "Estado": "Activa", "CUIT": "30711435227", "Nombre en AFIP": "ENERGIA ARGENTINA S.A. - YPF S.A. - PROYECTO GNL ESCOBAR - UNION TRANSITORIA DE EMPRESAS",
                    -"Nombre Soc SAP": "UTE RINCON DEL MANGRULLO", "Código SAP": "1160", "Estado": "Activa", "CUIT": "30714428469", "Nombre en AFIP": "YPF S.A - PAMPA ENERGIA S.A.. UNION TRANSITORIA DE EMPRESAS - RINCON DEL MANGRULLO",
                    -"Nombre Soc SAP": "UTE CHACHAHUEN", "Código SAP": "1164", "Estado": "Activa", "CUIT": "30716199025", "Nombre en AFIP": "YPF S.A.-KILWER S.A.-KETSAL S.A.-ENERGIA MENDOCINA S.A. AREA CHACHAHUEN UNION TRANSITORIA DE EMPRESAS",
                    -"Nombre Soc SAP": "UTE La amarga chica", "Código SAP": "1167", "Estado": "Activa", "CUIT": "30714869759", "Nombre en AFIP": "YPF S.A. - PETRONAS E&P ARGENTINA S.A.",
                    -"Nombre Soc SAP": "UTE EL OREJANO", "Código SAP": "1169", "Estado": "Activa", "CUIT": "30715142658", "Nombre en AFIP": "YPF S.A.- PBB POLISUR S.A., AREA EL OREJANO UNION TRANSITORIA",
                    -"Nombre Soc SAP": "CIA HIDROCARBURO NO CONVENCIONAL SRL", "Código SAP": "1171", "Estado": "Activa", "CUIT": "30714124427", "Nombre en AFIP": "COMPAÑIA DE HIDROCARBURO NO CONVENCIONAL S.R.L.",
                    -"Nombre Soc SAP": "UTE PAMPA (YSUR)", "Código SAP": "1632", "Estado": "Activa", "CUIT": "30711689067", "Nombre en AFIP": "APACHE ENERGIA ARGENTINA S.R.L. - PETROLERA PAMPA S.A., UNION TRANSITORIA DE EMPRESAS - ESTACION FERNANDEZ ORO Y ANTICLINAL CAMPAMENTO"
                    **Aclaración sobre lista de sociedades permitidas:**
                    - Cada elemento de la lista hace referencia a una unica sociedad.
                    - Cada apartado de un elemento sirve para identificar a la misma sociedad. Los apartados estan delimitados por ','.
                    - Si detectas un dato de la lista en el documento completa los datos del customer con los datos de la lista para ese customer.
                    - Cualquier nombre de sociedad o CUIT encontrado en el documento que no tenga match con la lista de sociedades deberá interpretarse como dato del Vendor.
                    - El campo "Signed" es un flag (booleano) para indicar si el documento está firmado. En caso de que reconozcas una firma deberás setear este campo como True.
                    """
            }
        ]
        user_content.append({
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{base64_data}"}
        })
        return user_content

    def parse_completion_response(self, completion):
        """
        Procesa la respuesta del modelo para extraer el JSON válido y convertirlo en un diccionario de campos.
        """
        extracted_data = completion.model_dump()
        content = extracted_data['choices'][0]['message']['content']
        json_content = content.strip('```json\n').strip('```')
        data = json.loads(json_content)

        return data

    def extract_fields(self, base64_images: list, fields_to_extract: List[str]):
        """
        Extrae datos específicos de una lista de imágenes en base64 y organiza los resultados en un diccionario.

        :param base64_images: Lista de diccionarios con datos de las imágenes (file_name y base64_content).
        :param fields_to_extract: Lista de campos a extraer.
        :return: Diccionario con los resultados extraídos o información de error.
        """
        try:
            if not base64_images or not isinstance(base64_images, list):
                raise ValueError("La lista de imágenes base64 no es válida.")
            if not fields_to_extract or not isinstance(fields_to_extract, list):
                raise ValueError("La lista de campos a extraer no es válida.")

            all_results = {}

            for index, image_data in enumerate(base64_images):
                file_name = image_data.get("file_name", f"unknown_{index + 1}")
                base64_content = image_data.get("base64_content", "")

                if not base64_content:
                    print(f"El archivo {file_name} no contiene datos base64.")
                    all_results[file_name] = {
                        "fields": {},
                        "missing_fields": [],
                        "error": "El contenido base64 está vacío.",
                    }
                    continue

                user_content = self.create_user_content(base64_content, fields_to_extract)

                messages = [
                    {"role": "system", "content": "Eres un asistente que extrae datos de documentos."},
                    {"role": "user", "content": user_content}
                ]

                try:
                    completion = self.openai_client.beta.chat.completions.parse(
                        model=self.gpt_model_name,
                        messages=messages,
                        max_tokens=5000,
                        temperature=0.1,
                        top_p=0.1,
                        logprobs=True,
                    )

                    prompt_tokens = completion.usage.prompt_tokens
                    completion_tokens = completion.usage.completion_tokens
                    total_tokens = prompt_tokens + completion_tokens
                    print(f"Tokens usados para {file_name}: {total_tokens} (Prompt: {prompt_tokens}, Completion: {completion_tokens})")

                    data = self.parse_completion_response(completion)

                    # Crear el diccionario de campos extraídos
                    extracted_fields = {field_name: data.get(field_name, None) for field_name in fields_to_extract}

                    # Identificar campos faltantes
                    missing_fields = [field for field, value in extracted_fields.items() if value is None]

                    # Guardar resultados en un diccionario
                    all_results[file_name] = {
                        "invoice_number": index + 1,
                        "fields": extracted_fields,
                        "missing_fields": missing_fields,
                        "error": "",
                        "source": "Vision",
                        "tokens": total_tokens
                    }

                except Exception as model_error:
                    print(f"Error procesando la imagen {file_name}: {model_error}")
                    all_results[file_name] = {
                        "fields": {},
                        "missing_fields": [],
                        "error": str(model_error),
                        "source": "Vision",
                        "tokens": total_tokens
                    }

            return all_results
        except Exception as e:
            print(f"Error general al extraer campos: {e}")
            return {"error": str(e)}





# Ejemplo de uso
# if __name__ == "__main__":
#     base64_files = [
#         {"file_name": "invoice1.pdf", "base64_content": "<base64_string1>"},
#         {"file_name": "invoice2.pdf", "base64_content": "<base64_string2>"}
#     ]  # Reemplaza <base64_string1> y <base64_string2> con los datos base64 reales
#     fields_to_extract = ["InvoiceId", "CustomerName", "TotalAmount"]

#     results = process_base64_files(base64_files, fields_to_extract)
#     print(results)



import re
from typing import Annotated, TypedDict
from langchain_core.messages import HumanMessage
from langgraph.graph import StateGraph, START, END, MessagesState
from langgraph.prebuilt import ToolNode, tools_condition
from agentiacap.configs.llms import llm4o
from agentiacap.configs.classes import Input
from agentiacap.configs.Prompt_Template import categories, reflection_definition, classifier_definition
from agentiacap.tools.evaluar_contexto import evaluar_contexto

categories = "\n".join(categories)

classification = classifier_definition | llm4o
llm_with_tools = reflection_definition | llm4o.bind_tools([evaluar_contexto])

class OutputState(TypedDict):
    category:Annotated[str, ...]

def input_node(state:Input) -> MessagesState:
    return {
        "messages": [
            HumanMessage(
                content=f"""A continuación te dejo el siguiente mail para que lo categorices,\n
                Asunto: {state['asunto']}.\n
                Cuerpo: {state['cuerpo']}.\n
                Las categorias posibles son:\n
                {categories}
                Si te parece que no aplica ninguna o la información parece escasa, incompleta o ambigua entonces categorizalo como 'Otras consultas'."""
            )
    ]}

async def classifier_node(state:MessagesState) -> MessagesState:
    result = await classification.ainvoke(state["messages"])
    return {"messages": [HumanMessage(content=result.content)]}

async def reflection_node(state: MessagesState) -> MessagesState:
    prompt = HumanMessage(
            content="""¿Es la categoría asignada coherente con el contexto del email? Para validar esto utilizá la tool 'evaluar contexto'.
            """
        )
    response = llm_with_tools.invoke(state["messages"]+[prompt])
    return {"messages": state["messages"]+[response]}

def output_node(state: MessagesState) -> OutputState:
    match = re.search(r"APROBADA:\s*\"([^\"]+)\"", state["messages"][-1].content)
    if match:
        categoria = match.group(1)  # Extraer el valor después de "APROBADA:"
        return {"category":categoria}
    return{"category":"Otras consultas"}  # Valor por defecto si no se logró aprobar la categoría.

# Defino edges
def should_continue(state: MessagesState) -> str:
    if "APROBADA" in state["messages"][-1].content:
        return "output"  # Si está aprobada, avanza al nodo output
    else:
        return "classifier"  # Si está rechazada, regresa al clasificador

# Defino grafo
builder = StateGraph(MessagesState, input=Input, output=OutputState)

builder.add_node("input", input_node)
builder.add_node("classifier", classifier_node)
builder.add_node("reflect", reflection_node)
builder.add_node("tools", ToolNode([evaluar_contexto]))
builder.add_node("output", output_node)

builder.add_edge(START, "input")
builder.add_edge("input", "classifier")
builder.add_edge("classifier", "reflect")
builder.add_conditional_edges("reflect", tools_condition, {"tools":"tools", END:"output"})
builder.add_conditional_edges("tools", should_continue, ["output", "classifier"])
builder.add_edge("output", END)

# Instancio graph
classifier = builder.compile()